/****************************************************************************** 
 *                      Sources BCB - www.developpez.com                      * 
 * -------------------------------------------------------------------------- * 
 *                                                                            *
 * HCDFunctions.cpp                                                           *
 * Version: 1.0.3 mise  jour le 13 fvrier 2005                              *
 * Copyright Cryptonyx (http://five.chez.tiscali.fr)                          * 
 *                                                                            * 
 * Test avec:                                                                * 
 * SE: Win98, WinXP                                                           *
 * IDE: BCB5                                                                  * 
 *                                                                            * 
 * -------------------------------------------------------------------------- * 
 *                                                                            *
 * Ce code illustre une compression/dcompression de type Huffman             *
 *                                                                            *
 * -------------------------------------------------------------------------- * 
 *                                                                            * 
 * Retrouvez d'autres ressources BCB sur :                                    * 
 *  - http://c.developpez.com/sources/bcb/                                    * 
 *  - http://c.developpez.com/faq/bcb/                                        * 
 * Ainsi que sur les forums (http://www.developpez.net/forums/)               * 
 *                                                                            * 
 ******************************************************************************/

#include <SysUtils.hpp>
#include <stdio.h>

#include "HCDFunctions.h"
#include "HCDMessagesForFunctions.h"
//---------------------------------------------------------------------------
// Fonctions de la classe THuffman
//---------------------------------------------------------------------------
__fastcall THuffman::THuffman()
{
}

//---------------------------------------------------------------------------
//
// Initialize
//
//---------------------------------------------------------------------------
bool THuffman::Initialize(const AnsiString InputFile, const AnsiString OutputFile)
{
    if(InputFile == "" || OutputFile == "") {
    	ErrorCode   = eBadFileName;
        ErrorString = ListOfErrorMessages[HCD_INCORRECT_FILE_NAME];
        return false;
    }
    Source = InputFile;
    Destination = OutputFile;
    // Ouverture du fichier d'entre
    if((SourceFileHandle = FileOpen(InputFile, fmOpenRead)) <= 0) {
    	ErrorCode = eOpenReadFailed;
        ErrorString = ListOfErrorMessages[HCD_OPEN_FILE_FAILURE];
        return false;
    }
    // Ouverture du fichier de sortie
    if((DestinationFileHandle = FileCreate(OutputFile)) < 0) {
        ErrorCode = eCreateFailed;
        ErrorString = ListOfErrorMessages[3];
        return false;
    }
    InputFileBuffer = (char *)NULL;
    OutputFileBuffer = (char *)NULL;
    TreeRoot = (THuffmanNode *)NULL;
    Statistics = (THuffmanStatistics *)NULL;
    ExpandedSize = CompressedSize = 0;
    NumberOfElements = 0;
    EndOfHeader = 0;
    ErrorCode = eCorrect;
    ErrorString = "";
    return true;
}

//---------------------------------------------------------------------------
//
// Reset
//
//---------------------------------------------------------------------------
void THuffman::Reset()
{
	//Suppression des donnes alloues dynamiquement
    if(Statistics)
    	delete [] Statistics;
    if(TreeRoot != (THuffmanNode *)NULL)
        TreeDelete(TreeRoot);
    if(SourceFileHandle)
        FileClose(SourceFileHandle);
    if(DestinationFileHandle)
        FileClose(DestinationFileHandle);
    SourceFileHandle = DestinationFileHandle = 0;
    if(InputFileBuffer)
        delete [] InputFileBuffer;
    if(OutputFileBuffer)
        delete [] OutputFileBuffer;
}

//---------------------------------------------------------------------------
//
//  FileAnalyze
//
//---------------------------------------------------------------------------
bool THuffman::FileAnalyze()
{
unsigned long int i;

    // Allocation du tableau de statistiques
    Statistics = new THuffmanStatistics[MAXIMUM_ELEMENTS];
    for(i = 0 ; i < MAXIMUM_ELEMENTS ; i++) {
        Statistics[i].Frequency = 0;
        Statistics[i].Length = 0;
    }
    // Alloue un tampon de la taille du fichier et y charge son contenu 
    ExpandedSize = FileSeek(SourceFileHandle, 0, 2);
    FileSeek(SourceFileHandle, 0, 0);           // revient au dbut du fichier
    if(!ExpandedSize) {
        ErrorCode = eEmptyFile;
        ErrorString = ListOfErrorMessages[HCD_NO_COMPRESSION];
        return false;
    }
    InputFileBuffer = new char[ExpandedSize + 1];
    if((ExpandedSize = (unsigned long int)FileRead(SourceFileHandle, InputFileBuffer, ExpandedSize)) != ExpandedSize) {
        ErrorCode = eReadFailed;
        ErrorString = ListOfErrorMessages[HCD_READ_FILE_FAILURE];
        return false;
    }
    // Parcours du tampon caractre par caractre - comptage
    for(i = 0; i < ExpandedSize; i++)
        Statistics[(unsigned char)InputFileBuffer[i]].Frequency++;

    ErrorCode = eCorrect;
    return true;
}

//---------------------------------------------------------------------------
//
//	Compresser
//
//	entre : nom du fichier d'entr, nom du fichier de sortie
//	sortie : vrai ou faux
//
//---------------------------------------------------------------------------
bool THuffman::FileCompress(const AnsiString InputFile, const AnsiString OutputFile)
{
unsigned long int i, j;
unsigned char BitPosition = 0, htCodeLength;
unsigned short int htCodeValue;       // 2 octets

    if(Initialize(InputFile, OutputFile) == false)
        return false;

    if((FileAnalyze() == false) || (FrequencyTreeBuild() == false) || (HeaderGenerate() == false)) {
        Reset();
        return false;
    }

    // Chargement du fichier d'entre
    FileSeek(SourceFileHandle, 0, 0);
    if(!FileRead(SourceFileHandle, InputFileBuffer, ExpandedSize)) {
      ErrorCode = eReadFailed;
      ErrorString = ListOfErrorMessages[HCD_READ_FILE_FAILURE];
      Reset();
      return false;
    }
    // Parcours du tampon d'entre caractre par caractre
    j = (unsigned long int)EndOfHeader;
    for(i = 0; i < ExpandedSize; i++) {
       // insertion du code correspondant dans le buffer
       htCodeLength = Statistics[(unsigned char)(InputFileBuffer[i])].Length;
       htCodeValue  = Statistics[(unsigned char)(InputFileBuffer[i])].Code;
       // le code commence  gauche de htCodeValue. si le bit le plus  gauche vaut 1,
       // le bit en cours du tampon de sortie est mis  1, sinon il vaut dj zro grce au memset
       while(htCodeLength) {
          if(htCodeValue & 0x8000)
             OutputFileBuffer[j] |= (unsigned char)(0x01 << (7 - BitPosition));
          htCodeValue <<= 1;
          htCodeLength--;
          if(++BitPosition == 8) {    // aprs 8 bits, on passe au caractre suivant du buffer
             BitPosition = 0;
             j++;
          }
       }
    }
    OutputFileBuffer[++j] = '\0';
    FileSeek(DestinationFileHandle, 0, 2); // on se positionne  la fin de l'entte
    if(FileWrite(DestinationFileHandle, OutputFileBuffer, j) != (int)j)
       ErrorCode = eWriteFailed;
    // la taille totale est j
    CompressedSize = j;
    ErrorCode = eCorrect;
    // Pourcentage de compression
    if((double)CompressedSize / ExpandedSize <= 1.0)
       ErrorString = Destination + ListOfErrorMessages[HCD_FILE_COMPRESSED_1]
       + IntToStr(100 - (100 * CompressedSize / ExpandedSize)) + ListOfErrorMessages[HCD_FILE_COMPRESSED_2];
    else
       ErrorString = Destination + ListOfErrorMessages[HCD_UNDER_COMPRESSED]
       + IntToStr((int)((100 * ((double)CompressedSize) / ExpandedSize) - 100)) + ListOfErrorMessages[HCD_FILE_COMPRESSED_2];

    Reset();
    if(ErrorCode == eWriteFailed)
        return false;
    return true;
}

//---------------------------------------------------------------------------
//
//	FileDecompress
//
//	entre : noms des fichiers
//	sortie : fichier decompress, taille decompresse  jour
//
//---------------------------------------------------------------------------
bool THuffman::FileDecompress(const AnsiString InputFile, const AnsiString OutputFile)
{
THuffmanNode *htNodePointer;
unsigned short int k;
unsigned long int i, j = 0;
char Data;

    if(Initialize(InputFile, OutputFile) == false)
        return false;

    if(HeaderRead() == false || CodeTreeBuild() == false) {
        Reset();
        return false;
    }

    OutputFileBuffer = new char[ExpandedSize + 1];
    // Parcours du tampon caractre par caractre
    htNodePointer = TreeRoot;
    for(i = (unsigned long int)EndOfHeader; i < CompressedSize; i++) {
        // Rcupration d'un octet
        Data = InputFileBuffer[i];

        // Dcodage des caractres bit par bit
        for(k = 0; (k < 8) && (j < ExpandedSize); k++) {
            // on teste le bit de poids fort pour parcourir l'arbre
            if((Data & 0x80) == 0x80)
                htNodePointer = htNodePointer->Next;
            else
                htNodePointer = htNodePointer->Pred;

            if(htNodePointer == NULL) {
                // il n'y a qu'un seul type d'lment dans le fichier d'origine,
                // ou bien l'arbre a t gnr de manire incorrecte
                if(NumberOfElements == 1) { // on pourrait traiter ce cas particulier autrement
                   OutputFileBuffer[j++] = (unsigned char)TreeRoot->Value;
                   // retour  la racine de l'arbre
                   htNodePointer = TreeRoot;
                }
                else {
                   ErrorCode = eBadTree;
                   ErrorString = ListOfErrorMessages[HCD_READ_TREE_FAILURE];
                   Reset();
                   return false;
                }
            }

            if(!(htNodePointer->Next || htNodePointer->Pred)) {
                // Le noeud est une feuille - criture de l'octet
                OutputFileBuffer[j++] = (unsigned char)htNodePointer->Value;
                // retour  la racine de l'arbre
                htNodePointer = TreeRoot;
            }
            Data <<= 1; // on passe au bit suivant
        }
    }
    ErrorCode = eCorrect;
    ErrorString = ListOfErrorMessages[HCD_FILE_DECOMPRESSED_1] + IntToStr(ExpandedSize) + ListOfErrorMessages[HCD_FILE_DECOMPRESSED_2];
    OutputFileBuffer[j] = '\0';
    if(FileWrite(DestinationFileHandle, OutputFileBuffer, j) != (int)j) {
        ErrorString = ListOfErrorMessages[HCD_WRITE_FILE_FAILURE];
        ErrorCode = eWriteFailed;
    }
    Reset();
    if(ErrorCode != eCorrect)
        return false;
    return true;
}

//---------------------------------------------------------------------------
//
//  HeaderGenerate
//
//---------------------------------------------------------------------------
bool THuffman::HeaderGenerate()
{
LongOrChar FileSize;
ShortOrChar ShortCode;
AnsiString FileName;
unsigned long int i, j;
int DelimiterPosition;

    // Allocation du tampon de sortie et mise  zro indispensable de tous ses bits
    // Le tampon de sortie doit tre plus grand que le tampon d'entre pour les fichiers
    // de petite taille  cause de l'adjonction de la table de conversion
    // On ajoute 1400 octets par mesure de scurit - cf. notes d'accompagnement
    OutputFileBuffer = new char[ExpandedSize + 1401];   // 1400 + 1
    memset(OutputFileBuffer, 0, ExpandedSize + 1400);

    // On saute 2 octets pour sauvegarder la taille de l'entte
    // Ecriture du numro de version
    ShortCode.ShortValue = (unsigned short int)VERSION;
    OutputFileBuffer[2] = ShortCode.CharSize[0];
    OutputFileBuffer[3] = ShortCode.CharSize[1];

    // Ecriture de la taille du fichier en octets
    FileSize.LongSize = ExpandedSize;
    for(i = 4; i < 8; i++)
        OutputFileBuffer[i] = FileSize.CharSize[i - 4];

    // Ecriture du nom du fichier original sans le chemin
    DelimiterPosition = Source.LastDelimiter(PATH_DELIMITER);
    if(DelimiterPosition < 1)
        // Il n'y a pas le chemin
        FileName = Source;
    else
        FileName = Source.SubString(DelimiterPosition + 1, Source.Length() - DelimiterPosition);

    // Ecriture de la taille du nom du fichier original - on coupe  255 caractres
    if((DelimiterPosition = FileName.Length()) > 255)
        DelimiterPosition = 255;
    OutputFileBuffer[i++] = (char)DelimiterPosition;
    for(j = 1; j <= (unsigned long int)DelimiterPosition; j++)
        OutputFileBuffer[i++] = FileName[j];

    // Ecriture du nombre d'lments diffrents
    ShortCode.ShortValue = NumberOfElements;
    OutputFileBuffer[i++] = ShortCode.CharSize[0];
    OutputFileBuffer[i++] = ShortCode.CharSize[1];

    for(j = 0; j < MAXIMUM_ELEMENTS; j++) {
        // Si le caractre a une frquence nulle on passe au suivant
        if(!Statistics[j].Frequency)
            continue;

        // Ecriture du caractre
        OutputFileBuffer[i++] = (unsigned char)j;

        // Ecriture du nombre de bits du caractre
        OutputFileBuffer[i++] = Statistics[j].Length;

        // Ecriture du code du caractre
        ShortCode.ShortValue = Statistics[j].Code;
        OutputFileBuffer[i++] = ShortCode.CharSize[1];
        if(Statistics[j].Length > 8)
            OutputFileBuffer[i++] = ShortCode.CharSize[0];
    }
    EndOfHeader = (unsigned short int)i;
    // Ecriture de la taille de l'entte
    ShortCode.ShortValue = EndOfHeader;
    OutputFileBuffer[0] = ShortCode.CharSize[0];
    OutputFileBuffer[1] = ShortCode.CharSize[1];
    ErrorCode = eCorrect;
    return true;
}

//---------------------------------------------------------------------------
//
//	HeaderRead
//
//---------------------------------------------------------------------------
bool THuffman::HeaderRead()
{
LongOrChar FileSize;
ShortOrChar ShortCode;
int i, CurrentElement;
int NumberOfBits;
unsigned short int j;
char pszShortBuffer[256];

    if(!SourceFileHandle) {
        ErrorCode = eOpenReadFailed;
        ErrorString = ListOfErrorMessages[HCD_OPEN_FILE_FAILURE];
        Reset();
        return false;
    }

    CompressedSize = FileSeek(SourceFileHandle, 0, 2);
    FileSeek(SourceFileHandle, 0, 0);
    InputFileBuffer = new char[CompressedSize + 1];
    if(FileRead(SourceFileHandle, InputFileBuffer, CompressedSize) != (int)CompressedSize) {
        ErrorCode   = eBadHeader;
        ErrorString = ListOfErrorMessages[HCD_READ_FILE_FAILURE];
        return false;
    }

    Statistics = new THuffmanStatistics[MAXIMUM_ELEMENTS];
    for(i = 0 ; i < MAXIMUM_ELEMENTS; i++) {
        Statistics[i].Code   = 0;
        Statistics[i].Length = 0;
    }

    // Lecture de la taille de l'entte pour contrle
    ShortCode.CharSize[0] = InputFileBuffer[0];
    ShortCode.CharSize[1] = InputFileBuffer[1];
    EndOfHeader = ShortCode.ShortValue;

    // Lecture du numro de version
    ShortCode.CharSize[0] = InputFileBuffer[2];
    ShortCode.CharSize[1] = InputFileBuffer[3];
    if(ShortCode.ShortValue < VERSION) {
        // Mauvaise version - pas d'action spcifique dans cette version
        ErrorCode   = eBadVersion;
        ErrorString = ListOfErrorMessages[HCD_BAD_HEADER_VERSION];
        return false;
    }

    // Lecture de la taille du fichier en octets
    // 'i' sert  lire les caractres chargs dans InputFileBuffer
    // et ne doit plus tre modifi autrement que par '++'
    for(i = 4; i < 8; i++)
        FileSize.CharSize[i - 4] = InputFileBuffer[i];
    ExpandedSize = FileSize.LongSize;

    // Lecture de la taille du nom du fichier original
    ShortCode.ShortValue = (unsigned short int)InputFileBuffer[i++];

    for(j = 0; j < ShortCode.ShortValue; j++)
        pszShortBuffer[j] = InputFileBuffer[i++];

    // Zero terminal
    pszShortBuffer[ShortCode.ShortValue] = 0;
    if(Destination == "")
        Destination.sprintf("%s", pszShortBuffer);

    // Lecture du nombre d'lments diffrents
    ShortCode.CharSize[0] = InputFileBuffer[i++];
    ShortCode.CharSize[1] = InputFileBuffer[i++];
    NumberOfElements      = ShortCode.ShortValue;

    for(j = 0; j < NumberOfElements; j++) {
        // Lecture du code du caractre
        CurrentElement = (unsigned char)InputFileBuffer[i++];
        NumberOfBits = (int)InputFileBuffer[i++];
        Statistics[CurrentElement].Length = (unsigned char)NumberOfBits;

        ShortCode.CharSize[1] = InputFileBuffer[i++];
        if(NumberOfBits <= 8)
            ShortCode.CharSize[0] = 0;
        else    // Lecture des 8 autres bits du code
            ShortCode.CharSize[0] = InputFileBuffer[i++];

        Statistics[CurrentElement].Code = ShortCode.ShortValue;
    }
    ErrorCode = eCorrect;
    ErrorString = "";
    return true;
}

//---------------------------------------------------------------------------
//
//  FrequencyTreeBuild
//
//---------------------------------------------------------------------------
bool THuffman::FrequencyTreeBuild()
{
THuffmanNode *htRoot, *htNode;
int i;

    if(!Statistics) {
        ErrorCode = eAnalyzeFailed;
        ErrorString = ListOfErrorMessages[HCD_ANALYSIS_FAILURE];
        return false;
    }

    // Cration de la liste initiale
    NumberOfElements = 0;
    htRoot = NULL;
    for(i = 0; i < MAXIMUM_ELEMENTS; i++) {
        // On ne considre que les caractres qui ont au moins une occurence
        if(Statistics[i].Frequency) {
            NumberOfElements++;
            htNode = new THuffmanNode();
            htNode->Value = (unsigned char) i;
            htNode->Frequency = Statistics[i].Frequency;
            htNode->Pred = NULL;
            htNode->Next = NULL;
            htRoot = InsertNode(htRoot, htNode);
        }
    }
    // Cas d'un fichier vide
    if(NumberOfElements == 0) {
        ErrorCode = eEmptyFile;
        ErrorString = ListOfErrorMessages[HCD_NO_COMPRESSION];
        return false;
    }

    // Gnration de l'arbre et dtermination des codes de Huffman
    TreeRoot = TreeGenerate(htRoot, NumberOfElements);
    if(TreeRoot == (THuffmanNode *)NULL) {
        ErrorCode = eBadTree;
    	ErrorString = ListOfErrorMessages[HCD_READ_TREE_FAILURE];
        return false;
    }
    if(CodeListGenerate(TreeRoot, (unsigned short int)0, (unsigned char)0) == false) {
        ErrorCode = eBadList;
    	ErrorString = ListOfErrorMessages[HCD_WRITE_TREE_FAILURE];
        return false;
    }
    // Cas d'un fichier  un seul type d'lment
    if(NumberOfElements == 1) {
        Statistics[(unsigned char)(TreeRoot->Value)].Code   = 0x0000;	// le code vaut zro
        Statistics[(unsigned char)(TreeRoot->Value)].Length = 0x01;		// sa longueur vaut 1
    }

    ErrorCode = eCorrect;
    ErrorString = "";
    return true;
}

//---------------------------------------------------------------------------
//
//	InsertNode
//
//	entre
//		pointeur sur racine de l'arbre
//		noeud  ajouter
//	sortie
//		pointeur sur racine de l'arbre
//
//---------------------------------------------------------------------------
THuffmanNode *THuffman::InsertNode(THuffmanNode *htNode, THuffmanNode *Node)
{
    if(htNode == NULL) // soit premier noeud soit dernier noeud
        return Node;

    if(Node->Frequency > htNode->Frequency) {
        if(InsertNode(htNode->Next, Node) == Node)
            htNode->Next = Node;
        return htNode;
    }
    else { // insre Node au bon emplacement
        Node->Next = htNode;
        Node->Pred = htNode->Pred;
        htNode->Pred = Node;
        if(Node->Pred != NULL)
            Node->Pred->Next = Node;
        return Node;
    }
}

//---------------------------------------------------------------------------
//
//	TreeGenerate
//
//	entre
//		pointeur sur liste
//		nombre d'lments de la liste
//	sortie
//		pointeur sur arbre
//
//---------------------------------------------------------------------------
THuffmanNode *THuffman::TreeGenerate(THuffmanNode *htRoot, int htNodes)
{
THuffmanNode *htNodePointer;		// un noeud de travail pour le transfert des donnes
THuffmanNode **htNodePointerTable;	// tableau vide  instancier pour contenir des pointeurs sur des htNodes
int i, j;							// 2 index pour parcourir le tableau de pointeurs
int RemainingNodes;					// un compteur de noeuds restants dans le tableau au fur et  mesure de l'avancement

	// prconditions : l'arbre existe et le nombre de noeuds est compris entre 1 et 255 - erreurs gres par l'appelant
	// l'arbre doit exister
	if(htRoot == (THuffmanNode *)NULL)
		return (THuffmanNode *)NULL;

	// le nombre de noeuds doit tre compris entre 1 et MAXIMUM_ELEMENTS
	if((htNodes < 1) || (htNodes > MAXIMUM_ELEMENTS))
		return (THuffmanNode *)NULL;

	// cas particulier du noeud unique - la mthode de gnration ne s'applique pas
	if(htNodes == 1)
		return htRoot;

	// traitement des cas de 2  N noeuds
	// cas particulier de 2 noeuds - une optimisation consiste  traiter ce cas avant l'appel
	// cration d'un tableau de pointeur sur le noeuds
	htNodePointerTable = new THuffmanNode *[htNodes];
	if(htNodePointerTable == (THuffmanNode **)NULL)
		return (THuffmanNode *)NULL;

	// transfert des noeuds dans le tableau
	htNodePointer = htRoot;
	for(i = 0; i < htNodes; i++) {
		htNodePointerTable[i] = htNodePointer;
		htNodePointer->Pred = (THuffmanNode *)NULL;	// ce pointeur n'est plus ncessaire et peut induire en erreur
		htNodePointer = htNodePointer->Next;
		htNodePointerTable[i]->Next = (THuffmanNode *)NULL;	// ce pointeur n'est plus ncessaire et peut induire en erreur
	}

	// dmarrage de la cration de l'arbre
	RemainingNodes = htNodes;
	while(RemainingNodes > 2) {
		htNodePointer = new THuffmanNode();		// instanciation d'un nouveau noeud
		if(htNodePointer == (THuffmanNode *)NULL) {
			free(htNodePointerTable);
			return (THuffmanNode *)NULL;
		}
		htNodePointer->Frequency = htNodePointerTable[0]->Frequency + htNodePointerTable[1]->Frequency;
		htNodePointer->Pred = htNodePointerTable[0];
		htNodePointer->Next = htNodePointerTable[1];
		// on cherche l'emplacement du nouveau noeud
		j = 2;
		while((j < RemainingNodes) && (htNodePointer->Frequency > htNodePointerTable[j]->Frequency))
        	j++;
		// on dcale les noeuds de gauche de 2 rangs vers la gauche
		for(i = 2; i < j; i++)
			htNodePointerTable[i - 2] = htNodePointerTable[i];
		// on insre le nouveau noeud
		htNodePointerTable[j - 2] = htNodePointer;
		// on dcale les noeuds de droite de 1 rang vers la gauche
		for(i = j; i < RemainingNodes; i++)
			htNodePointerTable[i - 1] = htNodePointerTable[i];
		// on 'enlve' un noeud
		RemainingNodes--;
	}
	// restent 2 noeuds
	htRoot = new THuffmanNode();
	if(htRoot != (THuffmanNode *)NULL) {
        htRoot->Pred = htNodePointerTable[0];
        htRoot->Next = htNodePointerTable[1];
	}
	delete [] htNodePointerTable;
    return htRoot;
}

//---------------------------------------------------------------------------
//
//	CodeListGenerate
//
//	entre
//		pointeur sur arbre
//		code en cours de dtermination
//		longueur du code en bits
//	sortie
//		pointeur sur arbre
//
//---------------------------------------------------------------------------
bool THuffman::CodeListGenerate(THuffmanNode *htRoot, unsigned short int htCodeValue, unsigned char htCodeBitLength)
{
     // prcondition : l'arbre existe
     if(htRoot == (THuffmanNode *)NULL)
		return false;

    // seules les feuilles ont un Code non nul
    if((htRoot->Pred == (THuffmanNode *)NULL) && (htRoot->Next == (THuffmanNode *)NULL)) {
        // le code doit commencer  partir du bit de poids le plus fort
        htCodeValue <<= (16 - htCodeBitLength);
        Statistics[(htRoot->Value)].Code   = htCodeValue;
        Statistics[(htRoot->Value)].Length = htCodeBitLength;
    }
    else {
        CodeListGenerate(htRoot->Pred, (unsigned short int)(htCodeValue << 1), (unsigned char)(htCodeBitLength + 1));
        CodeListGenerate(htRoot->Next, (unsigned short int)((htCodeValue << 1) | 0x01), (unsigned char)(htCodeBitLength + 1));
    }
	return true;
}

//---------------------------------------------------------------------------
//
//	CodeTreeBuild
//
//---------------------------------------------------------------------------
bool THuffman::CodeTreeBuild()
{
    // Cration de la racine
    TreeRoot = new THuffmanNode();
    TreeRoot->Next = (THuffmanNode *)NULL;
    TreeRoot->Pred = (THuffmanNode *)NULL;
    // on cre une feuille pour chaque lment  frquence non nulle
    for(int i = 0; i <  MAXIMUM_ELEMENTS; i++)
        if(Statistics[i].Length)
            LeafAdd(TreeRoot, Statistics[i].Code, Statistics[i].Length, (unsigned char)i);
    return true;
}

//---------------------------------------------------------------------------
//
//	LeafAdd
//
//	entre
//		pointeur sur arbre
//		code en cours de dtermination
//		longueur du code en bits
//      valeur a assigner
//	sortie
//		pointeur sur arbre
//
//---------------------------------------------------------------------------
THuffmanNode *THuffman::LeafAdd(THuffmanNode *htRoot, unsigned short int htCodeValue,
                            unsigned char htCodeBitLength, unsigned char Value)
{
    if(htRoot == (THuffmanNode *)NULL) {
        htRoot = new THuffmanNode();
        htRoot->Next = (THuffmanNode *)NULL;
        htRoot->Pred = (THuffmanNode *)NULL;
        if(htCodeBitLength) {
            if((htCodeValue & 0x8000) == 0x8000)
                htRoot->Next = LeafAdd(NULL, (unsigned short int)(htCodeValue << 1),
                                            (unsigned char)(htCodeBitLength - 1), Value);
            else
                htRoot->Pred = LeafAdd(NULL, (unsigned short int)(htCodeValue << 1),
                                            (unsigned char)(htCodeBitLength - 1), Value);
        }
        else
            htRoot->Value = Value;
    }
    else {
        if((htCodeValue & 0x8000) == 0x8000)
            htRoot->Next = LeafAdd(htRoot->Next, (unsigned short int)(htCodeValue << 1),
                                            (unsigned char)(htCodeBitLength - 1), Value);
        else
            htRoot->Pred = LeafAdd(htRoot->Pred, (unsigned short int)(htCodeValue << 1),
                                            (unsigned char)(htCodeBitLength - 1), Value);
    }
    return htRoot;
}

//---------------------------------------------------------------------------
//
//	TreeDelete
//      libre les ressources occupes par l'arbre des codes
//	entre
//		pointeur sur arbre
//
//---------------------------------------------------------------------------
void THuffman::TreeDelete(THuffmanNode *htRoot)
{
    if(htRoot == (THuffmanNode *)NULL)
        return;

    if((htRoot->Pred == (THuffmanNode *)NULL) && (htRoot->Next == (THuffmanNode *)NULL))
        delete htRoot;
    else {
        TreeDelete(htRoot->Pred);
        TreeDelete(htRoot->Next);
    }
}

//---------------------------------------------------------------------------
//
//  GetLastError : renvoie le dernier message d'erreur
//
//---------------------------------------------------------------------------
AnsiString THuffman::GetLastError()
{
    return ErrorString;
}

//---------------------------------------------------------------------------

