/******************************************************************************
 *                      Sources BCB - www.developpez.com                      *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * HCDFunctions.h                                                             *
 * Version: 1.0.4 mise  jour le 13 fvrier 2005                              *
 * Copyright Cryptonyx (http://five.chez.tiscali.fr)                          *
 *                                                                            *
 * Test avec:                                                                *
 * SE: Win98, WinXP                                                           *
 * IDE: BCB5                                                                  * 
 *                                                                            *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * Ce code illustre une compression/dcompression de type Huffman             *
 *                                                                            *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * Retrouvez d'autres ressources BCB sur :                                    *
 *  - http://c.developpez.com/sources/bcb/                                    *
 *  - http://c.developpez.com/faq/bcb/                                        *
 * Ainsi que sur les forums (http://www.developpez.net/forums/)               *
 *                                                                            *
 ******************************************************************************/
#ifndef HCDFunctionsH
#define HCDFunctionsH
//--------------------------------------------------------------------------
// Constantes
#define PATH_DELIMITER      "\\"
#define MAXIMUM_ELEMENTS    256
#define VERSION             0x10
//---------------------------------------------------------------------------
// Codes d'erreurs
typedef enum { eCorrect, eOpenReadFailed, eCreateFailed, eAnalyzeFailed, eEmptyFile, eWriteFailed,
	eReadFailed, eBadTree, eBadList, eBadVersion, eBadHeader, eBadFileName} THuffmanError;

//--------------------------------------------------------------------------
// Dfinition des structures de donnes
//--------------------------------------------------------------------------
// Structure ncessaires  la lecture et  l'criture des enttes
union long_or_char {
    unsigned long int LongSize;
    char CharSize[4];   // 4 octets pour un long int
};
typedef union long_or_char LongOrChar;
union short_or_char {
    unsigned short int ShortValue;
    char CharSize[2];   // 2 octets pour un short int
};
typedef union short_or_char ShortOrChar;

// structure d'un noeud ou d'une feuille de l'arbre de Huffman
struct THuffmanNode {
    unsigned char Value;        // Code ASCII
    unsigned int  Frequency;    // Frequence du code ou frquence cumule des sous-arbres
    THuffmanNode *Pred;         // Pointeur su prdecesseur
    THuffmanNode *Next;         // Pointeur sur suivant
};

// Structure des statistiques pour chaque lment d'analyse
struct THuffmanStatistics {
    unsigned long int  Frequency;
    unsigned short int Code;
    unsigned char      Length;      // Longueur du Code en bits
};

//--------------------------------------------------------------------------
// Classe Huffman permettant de raliser la (d)compression de Huffman
class THuffman
{
private :
    AnsiString Source;                      // Nom du fichier source
    AnsiString Destination;                 // Nom du fichier destination
    int SourceFileHandle;                   // Adresse du descripteur du fichier source
    int DestinationFileHandle;              // Adresse du descripteur du fichier de destination
    char *InputFileBuffer;                  // Tampon pour le fichier d'entre
    char *OutputFileBuffer;                 // Tampon pour le fichier de sortie
    THuffmanNode *TreeRoot;                 // Racine de l'arbre de Huffman
    unsigned short int NumberOfElements;    // Nombre d'lments diffrents du fichier  compresser
    unsigned long int ExpandedSize;         // Taille du fichier original
    unsigned long int CompressedSize;       // Taille du fichier compress avec entte
    unsigned short int EndOfHeader;         // Longueur de l'entte
    unsigned int ErrorCode;                 // Code de la dernire erreur survenue
    AnsiString ErrorString;                 // Message associ  la dernire erreur survenue
    THuffmanStatistics *Statistics;         // Statistiques du fichier source

    // Fonctions rcursives de bas niveau pour construire l'arbre et gnrer les codes
    THuffmanNode *InsertNode(THuffmanNode *Arbre, THuffmanNode *Node);
    THuffmanNode *TreeGenerate(THuffmanNode *htRoot, int htNodes);
    bool CodeListGenerate(THuffmanNode *htRoot, unsigned short int htCodeValue, unsigned char htCodeBitLength);
    THuffmanNode *LeafAdd(THuffmanNode *htRoot, unsigned short int htCodeValue, unsigned char htCodeBitLength, unsigned char Value);
    void TreeDelete(THuffmanNode *htRoot);

    // Fonctions ncessaires  la compression et  la dcompression
    bool FileAnalyze();             // Analyse le fichier source et alimente les statistiques
    bool HeaderGenerate();	        // Ecrit l'entte dans le fichier de destination
    bool HeaderRead();              // Lit l'entte  partir du fichier source
    bool FrequencyTreeBuild();      // Construit l'arbre de Huffman  partir des frequences ==> dduire le code
    bool CodeTreeBuild();           // Construit l'arbre de Huffman  partir des codes ==> dcompression
    bool Initialize(const AnsiString InputFile, const AnsiString OutputFile);
    void Reset();

public :
    // Fonctions d'interface de la classe
    bool FileCompress(const AnsiString InputFile, const AnsiString OutputFile);      // Compresse un fichier
    bool FileDecompress(const AnsiString InputFile, const AnsiString OutputFile);    // Dcompresse un fichier
    AnsiString GetLastError();
    __fastcall THuffman();
};
#endif
